#ifndef RETURN_CODES_H_
#define RETURN_CODES_H_

#define PYION_OK 0
#define PYION_SDR_ERR -10000
#define PYION_IO_ERR -10001
#define PYION_INTERRUPTED_ERR -10002
#define PYION_CONN_ABORTED_ERR -10003
#define PYION_MALLOC_ERR  -10004



#endif